;; Generated package description from peg.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "peg" "1.0.2.0.20251217.0" "Parsing Expression Grammars in Emacs Lisp" '((emacs "25")) :commit "2bccc414a94f067eb571fe614270494750a5de0e" :url "https://elpa.gnu.org/nongnu/peg.html" :authors '(("Helmut Eller" . "eller.helmut@gmail.com")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
